# -*- coding: future_fstrings -*-
"""Module for searching torrents on Distrowatch."""
import asyncio
import logging
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin


class Distrowatch(trackerplugin.TrackerPlugin):
    """Represent Distrowatch search engine."""

    log = logging.getLogger(__name__)

    filters = {"legal"}
    search_url = "https://distrowatch.com/dwres.php?resource=bittorrent"

    async def results(self, title: str) -> Iterator[torrent.Torrent]:
        """Get of list of torrent page urls.

        Args:
            title: String to search for.
        """
        try:
            page = await self.get(self.search_url, headers=self.user_agent)
        except (OSError, asyncio.TimeoutError):
            self.log.error("Cannot reach server at %s", self.search_url)
            return iter([])
        return (t for t in self._parse(page) if title in t.name)

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:
        """Parse result page.

        Args:
          raw: Raw HTML page.

        Returns:
            List of torrent names with URLs.
        """
        soup = bs4.BeautifulSoup(raw, "lxml")
        try:
            content = soup.find("table", cellpadding="5").find_all("tr")[1:]
            for line in content:
                info = line.select("td.torrent")[1]
                name = str(info.a.string.lower())
                url_stub = info.a.get("href")
                url = f"https://distrowatch.com/{url_stub}"
                tracker = self.__class__.__name__
                date = str(line.select("td.torrentdate")[0].string)
                yield torrent.Torrent(name, tracker, torrent=url, date=date)
        except AttributeError:
            yield from ()



"""Handle application configuration."""
import atexit
import collections
import logging
import os
import pathlib
from typing import Dict, List, Union

import pkg_resources
import ruamel.yaml

from mariner import utils

Value = Union[str, int, Dict, List]


class Config(collections.abc.MutableMapping):
    """Class to hold config settings using yaml. Behaves like a dictionary."""

    default_directory = "~/.config/mariner/"
    default_config = pkg_resources.resource_filename(__name__, "config/config.yaml")
    log = logging.getLogger(__name__)

    def __init__(
        self,
        configpath: str = None,
        _parent: "Config" = None,
        _config: Union["Config", Dict] = None,
    ) -> None:
        self._configpath = configpath
        self._yaml = ruamel.yaml.YAML()
        self._parent = _parent

        if not self._parent:
            self._config = self.load()
        else:
            self._config = _config

        # Save the config file on exit
        atexit.register(self.save)

    @property
    def configpath(self) -> pathlib.Path:
        """Create configuration file if necessary and return the path.

        Returns:
            Path to configuration file.
        """
        if self._configpath:
            path = pathlib.Path(self._configpath)
        else:
            directory = os.getenv("XDG_CONFIG_HOME", self.default_directory)
            directory = utils.check_path(directory)
            path = pathlib.Path(directory, "config.yaml")
        self.log.debug("path=%s", path)
        return path

    def load(self) -> Dict:
        """Load configuration saved in given path.

        Returns:
            Dictionary of configuration values.
        """
        try:
            raw_config = self.configpath.read_text()
        except FileNotFoundError:
            default_path = pathlib.Path(self.default_config)
            raw_config = default_path.read_text()
        return self._yaml.load(raw_config)

    def save(self) -> None:
        """Save the configuration."""
        if self._parent:
            self._parent.save()
        else:
            with self.configpath.open("w", encoding="utf-8") as file_:
                self.log.debug("configpath=%s file=%s", self.configpath, file_)
                self._yaml.dump(self._config, file_)

    def _as_config(self, dict_: Union[str, int, List, Dict]) -> Union[str, int, List, "Config"]:
        """Save config inside config.

        Args:
          dict_: Piece of configuration.

        Returns:
            Configuration as a dictionary.
        """
        if isinstance(dict_, collections.abc.MutableMapping):
            return Config(_parent=self, _config=dict_)
        return dict_

    def __getitem__(self, item: str) -> Value:
        if item not in self._config:
            raise KeyError(item)
        return self._config[item]

    def __setitem__(self, key: str, value: Value) -> None:
        self._config[key] = self._as_config(value)
        self.save()

    def __getattr__(self, attr):
        return self.__getitem__(attr)

    def __setattr__(self, attr: str, value: Value) -> None:
        if attr.startswith("_"):
            self.__dict__[attr] = value
        else:
            self.__setitem__(attr, value)

    def __delitem__(self, key: str) -> None:
        del self._config[key]

    def __iter__(self):
        for item in self._config:
            yield item

    def __len__(self) -> int:
        return len(self._config)

    def __repr__(self) -> str:
        return repr(self._config)

    def __str__(self) -> str:
        return str(self.__repr__())



"""Handle PirateBay alternative proxies."""
import bs4

from mariner import exceptions, proxyplugin


class PirateBayProxy(proxyplugin.ProxyPlugin):
    """PirateBay proxy list."""

    default_proxy = "https://thepiratebay.org"
    proxy_page_url = "https://proxybay.github.io"

    @staticmethod
    def _parse(page: str) -> str:
        """Parse the page and return a list of pages.

        Args:
            page: Proxy list page to parse.

        Returns:
            Urls of proxy websites.
        """
        soup = bs4.BeautifulSoup(page, "lxml")
        try:
            data = soup.select("table")[0].select("tr")[1:]
        except IndexError:
            raise exceptions.NoProxyAvailable("Cannot get proxy list")

        for site in data:
            if PirateBayProxy._is_online(site):
                return site.select("td.site")[0].a.get("href")

        raise exceptions.NoProxyAvailable("No proxy available for PirateBay")

    @staticmethod
    def _is_online(soup: bs4.Tag) -> bool:
        """Check if proxy is online or not.

        Args:
            soup: BeautifulSoup for Proxy data.

        Returns:
            True if site is listed online."""
        status = soup.select("td.status")[0].img.get("src")
        return "up" in status



# -*- coding: future_fstrings -*-
"""Module for searching torrents on Archive."""
import asyncio
import logging
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin


class Archive(trackerplugin.TrackerPlugin):
    """Represents Archive search engine."""

    log = logging.getLogger(__name__)

    filters = {"legal"}
    search_url = "https://archive.org/details/feature_films?and[]={title}&sin="

    async def results(self, title: str) -> Iterator[torrent.Torrent]:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        try:
            search_url = self.search_url.format(title=title)
            page = await self.get(search_url, headers=self.user_agent, timeout=self.timeout)
        except (OSError, asyncio.TimeoutError):
            self.log.error("Cannot reach server at %s", search_url)
            return iter([])
        return (t for t in self._parse(page) if title in t.name.casefold())

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results to parse.

        Returns:
            List of torrent names with magnet links and URLs.
        """
        soup = bs4.BeautifulSoup(raw, "lxml")
        try:
            contents = soup.select("div.results")[0].select("div.item-ia")[1:]
            for content in contents:
                name = str(content.select("div.ttl")[0].string.strip())
                tracker = self.__class__.__name__

                url_stub = content.get("data-id")
                url = f"https://archive.org/download/{url_stub}/{url_stub}_archive.torrent"

                yield torrent.Torrent(name, tracker, torrent=url)
        except IndexError:
            yield from []



"""Download a list of files to given location."""
import asyncio
import logging
import pathlib
from typing import List, Tuple, Union, Generator

import aiofiles
import aiohttp

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass

from mariner import utils

Url = str
Path = Union[str, pathlib.Path]
File = Tuple[Url, str]


class Downloader:
    """Handle file downloads asynchronously."""

    log = logging.getLogger(__name__)

    def __init__(self, download_path: Path = "~/Downloads", timeout: int = 10) -> None:
        self.download_path = utils.check_path(download_path)
        self.timeout = timeout

    async def download_coroutine(self, session: aiohttp.ClientSession, url: Url, name: str) -> Path:
        """Download a single file and asynchronously save it to disk.

        Args:
            url: URL of file.
            name: Name of file.

        Returns:
            Path of downloaded file.
        """
        filename = self.download_path / name
        self.log.debug("filename=%s", filename)
        async with session.get(url, timeout=self.timeout) as response:
            # Cast filename to str as workaround for Python 3.5
            async with aiofiles.open(str(filename), "wb") as file_:
                while True:
                    chunk = await response.content.read(1024)
                    if not chunk:
                        self.log.debug("saved=%s", filename)
                        break
                    await file_.write(chunk)
        return filename

    async def _download_all(self, download_list: List[File]) -> Generator:
        """Run a loop to download all files in the list.

        Args:
            download_list: List of files to download.

        Return:
            List of downloaded files.
        """
        async with aiohttp.ClientSession() as session:
            tasks = asyncio.as_completed(
                [self.download_coroutine(session, url, name) for url, name in download_list]
            )
            return await asyncio.gather(*tasks)

    def download(self, download_list: List[File]) -> Generator:
        """Run a loop to download all files in the list.

        Args:
            download_list: List of files to download.

        Return:
            List of downloaded files.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._download_all(download_list))



# -*- coding: future_fstrings -*-
"""Utility functions for Mariner."""
import pathlib
from typing import Any, Union

import colorama


def check_path(path: Union[str, pathlib.Path]) -> pathlib.Path:
    """Check if path exists. If not, create it.

    Args:
      path: Path to check.

    Returns:
      Resulting path.
    """
    # Convert to Path object if needed
    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)
    path = path.expanduser()

    # If path points to a file, get the parent directory
    directory = path.parent if path.suffix else path

    # Create path if needed
    directory.mkdir(parents=True, exist_ok=True)

    return path


def color(string: Any, text_color: str) -> str:
    """Color string in given color.

    Arguments:
        string: String to color.
        color: Defaults to yellow. Color to use.

    Returns:
        Colored string.
    """
    colors = {}
    colors["blue"] = colorama.Fore.BLUE
    colors["cyan"] = colorama.Fore.CYAN
    colors["green"] = colorama.Fore.GREEN
    colors["magenta"] = colorama.Fore.MAGENTA
    colors["red"] = colorama.Fore.RED
    colors["yellow"] = colorama.Fore.YELLOW
    if text_color not in colors.keys():
        raise ValueError(f"{text_color} is not a supported color.")
    return "".join([colors[text_color], str(string), colorama.Style.RESET_ALL])


def cyan(string: Any) -> str:
    """Color string cyan."""
    return color(string, "cyan")


def green(string: Any) -> str:
    """Color string cyan."""
    return color(string, "green")


def magenta(string: Any) -> str:
    """Color string magenta."""
    return color(string, "magenta")


def red(string: Any) -> str:
    """Color string red."""
    return color(string, "red")


def yellow(string: Any) -> str:
    """Color string yellow."""
    return color(string, "yellow")



# -*- coding: future_fstrings -*-
"""Module for searching torrents on KickAssTorrents."""
import asyncio
import logging
from typing import Dict, Iterator

import aiohttp
import bs4

from mariner import torrent, trackerplugin


class KickAssTorrents(trackerplugin.TrackerPlugin):
    """Represents KickAssTorrents search engine."""

    log = logging.getLogger(__name__)

    aliases = ["kat"]
    search_url = "https://katcr.co/katsearch/page/1/{title}"

    async def get_cookie(self, url: str, cookie) -> Dict:
        """Get KickAssTorrents session ID cookie.

        Args:
            url: Search page to get the cookie from.

        Returns:
            KickAssTorrents session cookie.
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url, headers=self.user_agent, timeout=self.timeout, cookies=cookie
            ) as response:
                cookie = response.cookies.popitem()
                return {cookie[0]: cookie[1].value}

    async def results(self, title: str) -> Iterator[torrent.Torrent]:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        try:
            search_url = self.search_url.format(title=title)
            try:
                # Kat needs two reloads to give the right cookie now
                cookie = await self.get_cookie(search_url, None)
                cookie = await self.get_cookie(search_url, cookie)
            except KeyError:
                # Couldn't get cookie
                cookie = None
            page = await self.get(search_url, headers=self.user_agent, cookies=cookie)
        except (OSError, asyncio.TimeoutError):
            self.log.error("Cannot reach server at %s", search_url)
            return iter([])
        return self._parse(page)

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results page to parse.

        Returns:
            List of torrent names with magnet links.
        """
        soup = bs4.BeautifulSoup(raw, "lxml")
        contents = soup.find("table", class_="torrents_table")
        try:
            for line in contents.find_all("tr")[1:]:
                torrent_ = line.find("div", class_="torrents_table__torrent_name")
                name = str(
                    torrent_.find("a", class_="torrents_table__torrent_title").b.string.strip()
                )
                magnet = torrent_.find("a", {"title": "Torrent magnet link"})["href"]
                tracker = self.__class__.__name__
                size = line.find("td", {"data-title": "Size"}).get_text().strip()
                date = line.find("td", {"data-title": "Age"}).get_text()
                raw_seeds = line.find("td", {"data-title": "Seed"}).string
                seeds = self._parse_number(raw_seeds)
                raw_leeches = line.find("td", {"data-title": "Leech"}).string
                leeches = self._parse_number(raw_leeches)
                yield torrent.Torrent(
                    name, tracker, magnet=magnet, size=size, date=date, seeds=seeds, leeches=leeches
                )
        except AttributeError:
            self.log.debug("No results found")
            yield from []



# -*- coding: future_fstrings -*-
"""Mariner download command."""
import logging

from cliff import lister

from mariner import downloader, utils


class Download(lister.Lister):
    """Download torrent with given ID."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument("ID", nargs="+", help="ID of the torrent to download", type=int)
        parser.add_argument(
            "--path",
            "-p",
            nargs="?",
            help="Path to downloads torrent files to",
            default=self.app.config["download_path"],
        )
        return parser

    def take_action(self, parsed_args):
        """Download chosen torrent.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of downloaded torrents and the location where they were saved.
        """
        path = parsed_args.path
        torrents = []
        for tid in parsed_args.ID:
            torrent_ = self.app.engine.result(tid)
            self.log.debug("tid=%s torrent=%s", tid, torrent_)
            if torrent_.torrent:
                torrents.append(torrent_)
                self.log.debug("Torrent appended.")
                self.log.info(f"Downloading torrent {utils.yellow(torrent_.name)}.")
            else:
                self.log.warning(
                    utils.yellow(f"No torrent for {torrent_.name}. Use magnet link instead.")
                )

        download_list = ((t.torrent, t.filename) for t in torrents)
        self.log.debug("download_list=%s download_path=%s", download_list, path)
        torrent_downloader = downloader.Downloader(download_path=path)
        filelist = torrent_downloader.download(download_list)

        headers = ("Name", "Saved to")
        colored_headers = [utils.magenta(h) for h in headers]
        columns = zip((t.colored().name for t in torrents), filelist)
        return (colored_headers, columns)



# -*- coding: future_fstrings -*-
"""Module for searching torrents on LinuxTracker."""
import logging
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin


class LinuxTracker(trackerplugin.TrackerPlugin):
    """Represents LinuxTracker search engine."""

    log = logging.getLogger(__name__)

    filters = {"legal"}
    search_url = "https://linuxtracker.org/index.php?page=torrents&search={title}"

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results page to parse.

        Returns:
            List of torrent names with magnet links.
        """
        soup = bs4.BeautifulSoup(raw, "lxml")
        content = soup.find_all("table", {"class": "lista", "width": "100%"})
        for torrent_ in content[3].select("td.lista"):
            try:
                name = str(torrent_.font.a.string)
                tracker = self.__class__.__name__

                links = torrent_.find_all("td", {"align": "right"})[0]
                magnet = links.find_all("a")[0]["href"]
                stub = links.find_all("a")[1]["href"]
                url = f"http://linuxtracker.org/{stub}"

                details = torrent_.find_all("tr")
                date = details[0].get_text().strip().split()[2]
                size = " ".join(details[1].get_text().split()[1:])
                raw_seeds = details[2].get_text().split()[1]
                seeds = self._parse_number(raw_seeds)
                raw_leeches = details[3].get_text().split()[1]
                leeches = self._parse_number(raw_leeches)

                yield torrent.Torrent(
                    name,
                    tracker,
                    torrent=url,
                    magnet=magnet,
                    date=date,
                    size=size,
                    seeds=seeds,
                    leeches=leeches,
                )
            except AttributeError:
                pass



# -*- coding: future_fstrings -*-
"""Mariner search command."""
import logging

from cliff import lister

from mariner import torrent, utils


class Search(lister.Lister):
    """Search for torrents."""

    log = logging.getLogger(__name__)

    @staticmethod
    def _availability(torrent_: torrent.Torrent) -> str:
        """Show whether it is available as torrent, magnet link or both."""
        availability = []
        if torrent_.magnet:
            availability.append("Magnet link")
        if torrent_.torrent:
            availability.append("Torrent")
        return ", ".join(availability)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument("title", help="Title to search for")
        parser.add_argument(
            "--all", "-a", action="store_true", help="Search all available trackers"
        )
        parser.add_argument(
            "--anime", "-A", action="store_true", help="Search trackers with anime content only"
        )
        parser.add_argument(
            "--filter",
            "-F",
            choices=self.app.engine.filters,
            help="Filter trackers that should be searched",
            default=[],
        )
        parser.add_argument(
            "--legal", "-L", action="store_true", help="Search trackers with legal content only"
        )
        parser.add_argument(
            "--limit",
            "-l",
            nargs="?",
            default=self.app.config["results"],
            help="Limit the number of results shown. Default is 50.",
            type=int,
        )
        parser.add_argument("--newest", "-n", action="store_true", help="Sort results by newest")
        parser.add_argument(
            "--trackers",
            "-t",
            action="append",
            choices=self.app.engine.plugins,
            help="Trackers that should be searched",
            default=[],
        )
        return parser

    @staticmethod
    def _parse_filters(parsed_args):
        """Return a list of filters to use.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of filters to use.
        """
        filters = []
        if parsed_args.legal:
            filters.append("legal")
        if parsed_args.anime:
            filters.append("anime")
        if parsed_args.filter:
            filters.append(parsed_args.filter)
        return set(filters)

    def _parse_trackers(self, parsed_args):
        """Return a list of trackers to use.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of trackers to use.
        """
        filters = self._parse_filters(parsed_args)

        if parsed_args.all:
            # Use all trackers
            trackers = self.app.engine.plugins.keys()
        elif filters:
            # Filter trackers to chosen filters
            trackers = [
                t for t in self.app.engine.plugins if filters <= self.app.engine.plugins[t].filters
            ]
        else:
            # If default tracker is used as default argument, the user provided ones
            # are appended to it, instead of replacing it.
            if not parsed_args.trackers:
                parsed_args.trackers.append(self.app.config["default_tracker"])
            trackers = [t.lower() for t in set(parsed_args.trackers)]
        return trackers

    def take_action(self, parsed_args):
        """Search for a torrent.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of torrents found.
        """
        title = parsed_args.title
        limit = parsed_args.limit
        newest = parsed_args.newest

        trackers = self._parse_trackers(parsed_args)

        self.log.info(f"Searching for {utils.cyan(title)}.")
        self.log.debug("title=%s limit=%s trackers=%s", title, limit, trackers)
        results = self.app.engine.search(title, trackers, limit, sort_by_newest=newest)

        headers = ("ID", "Name", "Tracker", "Seeds", "Size", "Uploaded", "Available as")
        # Headers cannot be a generator, as it messes up alignment
        colored_headers = [utils.magenta(h) for h in headers]
        colored_results = ((tid, t.colored()) for tid, t in results)
        columns = (
            (tid, t.name, t.tracker, t.seeds, t.size, t.date, self._availability(t))
            for tid, t in colored_results
        )
        return (colored_headers, columns)



# -*- coding: future_fstrings -*-
"""Module for searching torrents on LimeTorrents."""
import logging
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin


class LimeTorrents(trackerplugin.TrackerPlugin):
    """Represents LimeTorrents search engine."""

    log = logging.getLogger(__name__)

    aliases = ["lime"]
    search_url = "https://limetorrents.info/search/all/{title}/seeds/1"

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results to parse.

        Returns:
            List of torrent names with magnet links and URLs.
        """
        soup = bs4.BeautifulSoup(raw, "lxml")
        try:
            contents = soup.select("table.table2")[0].select("tr")[1:]
            for content in contents:
                data = content.select("td")
                name = str(data[0].select("a")[1].string)
                tracker = self.__class__.__name__

                url = data[0].div.a.get("href")

                size = str(data[2].string)
                date = str(data[1].string.split("-")[0].strip())
                seeds = self._parse_number(data[3].string)
                leeches = self._parse_number(data[4].string)

                yield torrent.Torrent(
                    name, tracker, torrent=url, size=size, date=date, seeds=seeds, leeches=leeches
                )
        except IndexError:
            yield from []



"""Base class for adding torrent tracker support."""
import abc
import asyncio
import logging
from typing import Iterator

from mariner import exceptions, mixins, torrent, proxyplugin

Url = str
Page = str


class TrackerMeta(abc.ABCMeta, type):
    """Metaclass to check that Tracker plugins override search_url."""

    def __new__(cls, name, bases, namespace, **kwargs):
        if abc.ABC not in bases:
            if not namespace.get("search_url"):
                raise exceptions.PluginError("You must define search_url")
        return type.__new__(cls, name, bases, namespace)


class TrackerPlugin(mixins.RequestMixin, abc.ABC, metaclass=TrackerMeta):
    """Represent a search engine."""

    log = logging.getLogger(__name__)
    aliases = []  # Aliases for the tracker name
    filters = set()

    data = {}
    request_method = "get"
    search_url = ""  # To be overwritten by subclasses

    def __init__(self, timeout: int = 10) -> None:
        super().__init__()
        self.timeout = timeout

    async def results(self, title: str) -> Iterator:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        try:
            if self.data:
                self.data["keyword"] = self.data.get("keyword").format(title=title)
            search_url = self.search_url.format(title=title)
            page = await self.request(
                self.request_method, search_url, data=self.data, timeout=self.timeout
            )
        except (OSError, asyncio.TimeoutError) as e:
            print(e)
            self.log.error("Cannot reach server at %s", search_url)
            return iter([])
        return self._parse(page)

    @abc.abstractmethod
    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:
        """Parse result page.

        Args:
            raw: Raw HTML page.

        Returns:
            List of torrents names with URLs and magnet links.

        """
        raise NotImplementedError

    @staticmethod
    def _parse_number(number: str) -> int:
        """Parse a number string from HTML page and return an interger.

        Args:
            number: Number string to parse.

        Return:
            Parsed number.
        """
        if number == "-":
            return 0
        squashed = number.replace(" ", "")
        return int(squashed.replace(",", ""))


class ProxyTrackerPlugin(mixins.RequestMixin, abc.ABC, metaclass=TrackerMeta):
    """Base class for trackers, that support alternative proxies.

    Attributes:
        proxies: Override with instance of ProxyPlugin.
    """

    log = logging.getLogger(__name__)
    aliases = []  # Aliases for the tracker name
    filters = set()
    request_method = "get"

    search_url = ""  # To be overwritten by subclasses

    def __init__(self, timeout: int = 10) -> None:
        super().__init__()
        self.proxies = proxyplugin.ProxyPlugin()
        self.timeout = timeout

    async def get_proxy(self) -> str:
        """Return a responding proxy.

        Returns:
            URL for responding proxy.
        """
        return await self.proxies.get_proxy()

    async def results(self, title: str) -> Iterator:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        proxy = await self.get_proxy()
        search_url = self.search_url.format(proxy=proxy, title=title)
        try:
            page = await self.request(self.request_method, search_url, timeout=self.timeout)
        except (OSError, asyncio.TimeoutError) as e:
            print(e)
            self.log.error("Cannot reach server at %s", search_url)
            return iter([])
        return self._parse(page)

    @abc.abstractmethod
    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:
        """Parse result page.

        Args:
            raw: Raw HTML page.

        Returns:
            List of torrents names with URLs and magnet links.

        """
        raise NotImplementedError

    @staticmethod
    def _parse_number(number: str) -> int:
        """Parse a number string from HTML page and return an interger.

        Args:
            number: Number string to parse.

        Return:
            Parsed number.
        """
        squashed = number.replace(" ", "")
        return int(squashed.replace(",", ""))

    @staticmethod
    def _parse_date(date: str) -> str:
        """Parse exotic date format, that Maya can't handle.

        Args:
            date: Date string to parse.

        Return:
            Parsed date.
        """
        if date.casefold() == 'y-day':
            return 'yesterday'
        return date



"""Entrance point for Mariner application."""
import logging
import os
import pathlib
import sys
from typing import List

from cliff import app, commandmanager
import colorama

from mariner import __version__
from mariner import searchengine, utils, config


class Mariner(app.App):
    """Interactive CLI for Mariner."""

    log = logging.getLogger(__name__)

    def __init__(self):
        super().__init__(
            description=__version__.__description__,
            version=__version__.__version__,
            command_manager=commandmanager.CommandManager("mariner.cli"),
            deferred_help=True,
        )
        self.config = None
        self.engine = None

    def build_option_parser(self, description, version, argparse_kwargs=None):
        """Return an Argparse option parser for Mariner.

        Args:
            description: Mariner description.
            version: Mariner version.
        """
        # Override the help message for log output
        kwargs = {"conflict_handler": "resolve"}
        parser = super().build_option_parser(description, version, kwargs)
        parser.add_argument(
            "--log-file",
            action="store",
            default=None,
            help="Specify a file to log output. Default ~/.local/share/mariner/mariner.log.",
        )
        parser.add_argument(
            "--config-file",
            "-c",
            action="store",
            default=None,
            help="Path to config file. Default ~/.config/mariner/config.yaml",
        )
        return parser

    def configure_logging(self):
        """Create logging handlers for any log output."""
        root_logger = logging.getLogger("")
        root_logger.setLevel(logging.DEBUG)

        # Set logging to file by default.
        if not self.options.log_file:
            log_dir = os.getenv("XDG_DATA_HOME", "~/.local/share/mariner")
            log_dir = utils.check_path(log_dir)
            self.options.log_file = str(pathlib.Path(log_dir) / "mariner.log")

        # Monkey patched to use RotatingFileHandler
        file_handler = logging.handlers.RotatingFileHandler(
            filename=self.options.log_file, maxBytes=1000000, backupCount=1
        )
        formatter = logging.Formatter(self.LOG_FILE_MESSAGE_FORMAT)
        file_handler.setFormatter(formatter)
        root_logger.addHandler(file_handler)

        # Always send higher-level messages to the console via stderr
        console = logging.StreamHandler(self.stderr)
        console_level = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}.get(
            self.options.verbose_level, logging.DEBUG
        )
        console.setLevel(console_level)
        formatter = logging.Formatter(self.CONSOLE_MESSAGE_FORMAT)
        console.setFormatter(formatter)
        root_logger.addHandler(console)

    def initialize_app(self, argv: List[str]) -> None:
        """Initialize the application.

        Args:
          argv: List[str]: List of command line arguments.
        """
        self.LOG.debug("Initialize Mariner")
        # Initialize color output
        colorama.init()

        if self.options.config_file:
            self.config = config.Config(configpath=self.options.config_file)
        else:
            self.config = config.Config()
        self.config.load()

        # Older configurations have no timeout option
        # Also, cast timeout to int, int case it was string
        try:
            timeout = int(self.config["timeout"])
        except KeyError:
            timeout = 10
            self.config["timeout"] = timeout

        self.engine = searchengine.SearchEngine(timeout=timeout)

        if self.interactive_mode:
            self.log.info("Welcome to Mariner. Navigate torrents with CLI!")
            self.log.info('Type "help" or see http://radek-sprta.gitlab.io/mariner to get started.')

    def prepare_to_run_command(self, cmd) -> None:
        """Code to run before command.

        Args:
          cmd: Command to run.
        """
        self.LOG.debug("Preparing to run command %s", cmd.__class__.__name__)

    def clean_up(self, cmd, result, err) -> None:
        """Code to run after command.

        Args:
          cmd: Command that was run.
          result: Result of the command.
          err: Errors caught while running the command.
        """
        self.LOG.debug("Clean up %s", cmd.__class__.__name__)
        if err:
            self.LOG.debug("Got an error: %s", err)


def main(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    """Application entrance point.

    Args:
      argv: Defaults to sys.argv[1:]. Command line argument list.
    """
    mariner = Mariner()
    return mariner.run(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# -*- coding: future_fstrings -*-
"""Represent a downloadable torrent."""
import datetime
from typing import Union

from mariner import mixins, utils


class Torrent(mixins.ComparableMixin):  # pylint: disable=too-many-instance-attributes
    """Class representing a Torrent."""

    def __init__(
        self,
        name: str,
        tracker: str,
        *,
        torrent: str = None,
        magnet: str = None,
        size: str = "Unknown",
        seeds: Union[int, str] = -1,
        leeches: Union[int, str] = None,
        date: Union[datetime.date, str] = datetime.date(1, 1, 1)
    ) -> None:
        self.name = name
        self.tracker = tracker
        self.torrent = torrent
        self.magnet = magnet
        self.size = size
        self.seeds = seeds
        self.leeches = leeches
        self.date = date

    @property
    def _cmpkey(self) -> int:
        """Key to use for torrent comparison."""
        return self.seeds

    @property
    def date(self) -> datetime.date:
        """Upload date in structured format."""
        return self._date

    @date.setter
    def date(self, value: str) -> None:
        if isinstance(value, (datetime.date, type(None))):
            self._date = value  # pylint: disable=attribute-defined-outside-init
        else:
            # Defer importing maya, as it is a slow import
            import maya # pylint: disable=import-outside-toplevel
            try:
                self._date = maya.when(value).date  # pylint: disable=attribute-defined-outside-init
            except ValueError:
                self._date = maya.parse(value).date  # pylint: disable=W0201
            except TypeError:
                self._date = datetime.date(1, 1, 1)  # pylint: disable=W0201

    @property
    def filename(self) -> str:
        """Filename to use for the torrent."""
        return self.name if ".torrent" in self.name else f"{self.name}.torrent"

    def colored(self) -> "Torrent":
        """Return Torrent with colored fields."""
        return Torrent(
            name=utils.yellow(self.name[:80]),
            tracker=self.tracker,
            torrent=self.torrent,
            magnet=self.magnet,
            size=self.size,
            seeds=utils.green(self.seeds),
            leeches=utils.red(self.leeches) if self.leeches is not None else None,
            date=self.date,
        )

    def __repr__(self) -> str:
        return f"Torrent({self.name}, {self.tracker}, {self.torrent}, \
{self.magnet}, {self.size}, {self.seeds}, {self.leeches}, {self.date})"

    def __str__(self) -> str:
        return self.__repr__()



# -*- coding: future_fstrings -*-
"""Mariner magnet command."""
import logging

from cliff import command
import pyperclip

from mariner import utils


class Magnet(command.Command):
    """Copy magnet link with given ID to clipboard."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument("ID", help="ID of the magnet link to copy", type=int)
        return parser

    def take_action(self, parsed_args):
        """Copy chosen magnet link to clipboard.

        Args:
            parsed_args: List of parsed arguments.
        """
        tid = parsed_args.ID
        torrent_ = self.app.engine.result(tid)
        self.log.debug("tid=%s torrent=%s", tid, torrent_)
        try:
            pyperclip.copy(torrent_.magnet)
            self.log.info(f"Copied {torrent_.colored().name} magnet link to clipboard.")
            self.log.debug("magnet=%s", torrent_.magnet)
        except pyperclip.PyperclipException:
            pyperclip.copy(torrent_.torrent)
            self.log.warning(utils.yellow(f"{torrent_.name} has no magnet link."))
            self.log.info(f"Copied {torrent_.colored().name} torrent URL to clipboard instead.")



# -*- coding: future_fstrings -*-
"""Module for searching torrents on TokyoTosho."""
import logging
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin


class TokyoTosho(trackerplugin.TrackerPlugin):
    """Represents TokyoTosho search engine."""

    log = logging.getLogger(__name__)

    aliases = ["tt"]
    filters = {"anime"}
    search_url = "https://www.tokyotosho.info/search.php?terms={title}"

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results to parse.

        Returns:
            List of torrent names with magnet links and URLs.
        """
        soup = bs4.BeautifulSoup(raw, "lxml")
        contents = soup.select("tr.category_0")
        for content in contents:
            try:
                # Even lines
                torrent_ = content.select(".desc-top")[0]
                links = torrent_.select("a")
                magnet = links[0].get("href")
                url = links[1].get("href")
                name = TokyoTosho._parse_name(links[1].contents)
                tracker = self.__class__.__name__
                result = torrent.Torrent(name, tracker, magnet=magnet, torrent=url)
            except IndexError:
                # Odd lines
                details = content.select("td.desc-bot")[0].get_text().split("|")
                result.size = details[1].split(":")[1].strip()
                result.date = details[2].split()[1]

                stats = content.select("td.stats")[0].select("span")
                raw_seeds = stats[0].string
                result.seeds = self._parse_number(raw_seeds)
                raw_leeches = stats[1].string
                result.leeches = self._parse_number(raw_leeches)
                yield result

    @staticmethod
    def _parse_name(raw: bs4.NavigableString) -> str:
        """Parse torrent name.

        Args:
          raw: Raw HTML name to parse.

        Returns:
            Parsed name.
        """
        name = [c for c in raw if isinstance(c, bs4.element.NavigableString)]
        return "".join(name)



"""Define exceptions used within the Mariner module."""


class Error(Exception):
    """Base - class for all exceptions raised by this module."""


class ConfigurationError(Error):
    """Illegal option in Configuration."""


class InputError(Error):
    """Wrong input by user."""


class NoProxyAvailable(Error):
    """No proxy available for given tracker."""


class NoResultException(Error):
    """No result found for given search string."""


class PluginError(Error):
    """Error in one of the plugins."""



# -*- coding: future_fstrings -*-
"""Module for searching torrents on PirateBay."""
import logging
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin
from mariner.proxies import piratebay


class PirateBay(trackerplugin.ProxyTrackerPlugin):
    """Represents PirateBay search engine."""

    log = logging.getLogger(__name__)

    aliases = ["tpb", "pb"]
    search_url = "{proxy}/search/{title}"

    def __init__(self, timeout: int = 10) -> None:
        super().__init__()
        self.proxies = piratebay.PirateBayProxy()

    async def get_proxy(self) -> str:
        """Return a responding proxy.

        Returns:
            URL for responding proxy.
        """
        return await self.proxies.get_proxy()

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results page to parse.

        Returns:
            List of torrent names with magnet links.
        """
        soup = bs4.BeautifulSoup(raw, "lxml")
        content = soup.find("table", id="searchResult")
        try:
            torrents = content.find_all("tr")
        except AttributeError:
            # No search result
            yield from []
        else:
            # First item is table header, last are links to other pages
            for torrent_ in torrents[1:-1]:
                raw_name = torrent_.find("a", class_="detLink")
                name = PirateBay._parse_name(raw_name)
                links = torrent_.find_all("a")
                magnet = links[3].get("href")
                tracker = self.__class__.__name__

                numbers = torrent_.find_all("td", align="right")
                raw_seeds = numbers[0].string
                seeds = self._parse_number(raw_seeds)
                raw_leeches = numbers[1].string
                leeches = self._parse_number(raw_leeches)

                description = torrent_.find("font", class_="detDesc").get_text()
                fields = description.split(",")
                date = self._parse_date(" ".join(fields[0].split()[1:]))
                size = " ".join(fields[1].split()[-2:])

                yield torrent.Torrent(
                    name, tracker, magnet=magnet, seeds=seeds, leeches=leeches, date=date, size=size
                )

    @staticmethod
    def _parse_name(raw: bs4.NavigableString) -> str:
        """Parse torrent name.

        Args:
          raw: Raw name to parse.

        Returns:
            Parsed name.
        """
        name = raw.string
        if name is None:
            name = raw["title"].split(" ")[2:]
            name = " ".join(str(x) for x in name)
        return name.strip()



"""Utility mixins used in Mariner."""
from typing import Dict, Union

import aiohttp


class ComparableMixin:  # pylint: disable=too-few-public-methods
    """Implement comparisons, using _cmpkey attribute."""

    def _compare(self, other, method):
        """Wrap rich comparison methods.

        Args:
            other: Object to compare to.
            method: Comparison method to use.

        Returns:
            Comparison function.
        """
        try:
            return method(self._cmpkey, other._cmpkey)  # pylint: disable=protected-access
        except (AttributeError, TypeError):
            # _cmpkey not implemented, or returned different type,
            # so cannot make a comparison with "other".
            return NotImplemented

    def __lt__(self, other):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s, o: s != o)


class RequestMixin:  # pylint: disable=too-few-public-methods
    """Mixin for HTTP requests."""

    user_agent = {
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:62.0) Gecko/20100101 Firefox/62.0"
    }

    async def get(
        self,
        url: str,
        *,
        headers: Dict[str, str] = None,
        cookies: Dict[str, str] = None,
        timeout: int = None
    ) -> str:
        """Get the requested page.

        Args:
            url: Url of the page to get.
            headers: Request headers.
            cookies: Request cookies.
            timeout: Request timeout.

        Returns:
            Raw HTML page.
        """
        return await self.request("get", url, headers=headers, cookies=cookies, timeout=timeout)

    async def request(
        self,
        method: str,
        url: str,
        *,
        data: Union[Dict, bytes] = None,
        headers: Dict[str, str] = None,
        cookies: Dict[str, str] = None,
        timeout: int = 10
    ) -> str:
        """Make a request.

        Args:
            method: Request method.
            url: Url of the page.
            data: Request payload.
            headers: Request headers.
            cookies: Request cookies.
            timeout: Request timeout.

        Returns:
            HTTP response.
        """
        headers = {**headers, **self.user_agent} if headers else self.user_agent

        async with aiohttp.ClientSession(headers=headers, cookies=cookies) as session:
            async with session.request(method, url, data=data, timeout=timeout) as response:
                return await response.text()



"""Mariner details command."""
import logging

from cliff import show

from mariner import utils


class Details(show.ShowOne):
    """Show details about torrent with given ID."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument("ID", help="ID of the torrent to show", type=int)
        return parser

    @staticmethod
    def _order_details(unordered):
        """Order results, so that Name is first.

        Args:
            unordered: Unordered results.

        Returns:
            ordered: Ordered results.
        """
        ordered = {}
        ordered["Name"] = unordered.pop("Name")
        ordered.update(unordered)
        # Shorten magnet link
        try:
            ordered["Magnet"] = ordered["Magnet"][:80] + "..."
        except KeyError:
            pass
        return ordered

    def take_action(self, parsed_args):
        """Show details for chosen torrent.

        Args:
            parsed_args: List of parsed arguments.

        Returns:
            List of details about chosen torrent.
        """
        tid = parsed_args.ID
        torrent_ = self.app.engine.result(tid)

        # Dictionary of attributes, that are not empty
        details = {
            d[0].strip("_").title(): d[1]
            for d in torrent_.colored().__dict__.items()
            if d[1] is not None
        }
        ordered_details = self._order_details(details)

        colored_keys = (utils.magenta(key) for key in ordered_details)
        return (colored_keys, ordered_details.values())



# -*- coding: future_fstrings -*-
"""Handle searching for torrents on torrent trackers."""
import asyncio
import importlib
import inspect
import itertools
import logging
import pathlib
from typing import List, Iterator, Optional, Set, Tuple, Union

import cachalot

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass

from mariner import exceptions, torrent

Name = str
Page = str
Path = Union[str, pathlib.Path]
Url = str


class SearchEngine:
    """Search on for a torrent using Tracker plugins."""

    log = logging.getLogger(__name__)
    plugin_directory = pathlib.Path(__file__).parent / "trackers"

    def __init__(self, timeout: int = 10) -> None:
        self.timeout = timeout
        self.plugins = {}
        self.results = cachalot.Cache(path="~/.local/share/mariner/results.json", size=1000)
        self.initialize_plugins()

    @property
    def filters(self) -> Set[str]:
        "Available filters."
        filters = set()
        for plugin in self.plugins.values():
            filters = filters.union(plugin.filters)
        return filters

    @property
    def timeout(self) -> int:
        "Timeout for connections."
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: int) -> None:
        if timeout <= 0:
            raise exceptions.ConfigurationError("Timeout cannot be negative.")
        self._timeout = timeout  # pylint: disable=attribute-defined-outside-init

    @staticmethod
    def _flatten(nested_list: List[List]) -> List:
        """Flatten a list."""
        return list(itertools.chain(*nested_list))

    def _load_modules(self) -> Iterator:
        """Find and import tracker plugins."""
        for module in self.plugin_directory.glob("*.py"):  # pylint: disable=no-member
            # Cast to str as a workaround for Python 3.5
            name = str(module.stem)
            module = str(module)

            self.log.debug("Loading module=%s", module)
            spec = importlib.util.spec_from_file_location(name, module)
            loaded_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(loaded_module)
            yield loaded_module

    def initialize_plugins(self) -> None:
        """Find engines and register them."""
        self.log.debug("Initializing plugins")
        for module in self._load_modules():
            classes = inspect.getmembers(module, inspect.isclass)
            for name, plugin in classes:
                # We don't care about classes defined in other modules
                if not plugin.__module__ == module.__name__:
                    continue
                self.log.debug("Adding plugin=%s", plugin)
                self.plugins[name.lower()] = plugin
                for alias in plugin.aliases:
                    self.plugins[alias] = plugin

    def result(self, tid: str) -> torrent.Torrent:
        """Get torrent of given id.

        Args:
            tid: ID of the torrent to get.

        Returns:
            Torrent with given ID.
        """
        self.log.debug("Fetching torrent with tid=%s", tid)
        torrent_ = self.results.get(tid)
        if torrent_:
            return torrent_
        raise exceptions.NoResultException(f"No torrent with ID {tid}")

    @cachalot.Cache(path="~/.cache/mariner/cache.json", timeout=14400, size=100)
    def _cached_search(self, title: str, trackers: List[str]) -> List[torrent.Torrent]:
        """Search for torrents on given site and cache to results. This
        method is an implementation detail. As coroutines are not easily
        serializable, we cannot simply cache TrackerPlugun.results() method.

        Args:
            title: String to search for.

        Returns:
            List of Torrents returned by the search.
        """
        self.log.debug("Fetching search results")

        # Get unique trackers
        try:
            plugins = set(self.plugins[t] for t in trackers)
        except KeyError:
            raise exceptions.ConfigurationError("Illegal value for default_tracker")
        else:
            tasks = asyncio.gather(
                *(p(timeout=self.timeout).results(title.lower()) for p in plugins)
            )
            loop = asyncio.get_event_loop()
            torrents = loop.run_until_complete(tasks)
            torrents = self._flatten(torrents)
        return torrents

    def search(
        self,  # pylint: disable=too-many-arguments
        title: str,
        trackers: List[str],
        limit: Optional[int] = 10,
        sort_by_newest: bool = False,
    ) -> List[Tuple[int, torrent.Torrent]]:
        """Search for torrents on given site.

        Args:
            title: String to search for.
            limit: Defaults to 10. Limits the number of results, that should be shown.

        Returns:
            List of Torrents returned by the search, up to the limit.
        """
        if not title:
            raise exceptions.InputError("No string to search for.")
        if not trackers:
            raise exceptions.InputError("No torrent trackers to search on")
        if limit <= 0:
            raise exceptions.InputError("Limit has to be higher than zero.")

        torrents = self._cached_search(title, trackers)
        sorted_torrents = self._sort_results(torrents, sort_by_newest)

        # Show only results up to to the limit
        results = list((i, t) for i, t in enumerate(sorted_torrents[:limit]))
        if results:
            self.save_results(results)
            return results
        raise exceptions.NoResultException(f"No results for {title}")

    def save_results(self, torrents: List[Tuple[int, torrent.Torrent]]) -> None:
        """Save results in a database.

        Args:
            torrents: List of ID, Torrent tuples.
        """
        self.results.clear()
        for tid, torrent_ in torrents:
            self.results.insert(tid, torrent_)

    def _sort_results(
        self, torrents: List[torrent.Torrent], sort_by_newest: bool
    ) -> List[torrent.Torrent]:
        """Sort torrent results.

        Args:
            torrents: List of torrents to sort.
            sort_by_newest: True if torrent should be sorted by recendivity.

        Returns:
            Sorted list of torrents.
        """
        self.log.debug("Sorting results")
        if sort_by_newest:
            sorted_torrents = list(reversed(sorted(torrents, key=lambda x: x.date)))
        else:
            sorted_torrents = list(reversed(sorted(torrents)))
        return sorted_torrents



# -*- coding: future_fstrings -*-
"""Module for declaring list of tracker proxies."""
import abc
import asyncio
import logging

import bs4

from mariner import exceptions, mixins

Url = str


class ProxyMeta(abc.ABCMeta, type):
    """Metaclass to check that ProxyList plugins override proxy_page_url."""

    def __new__(cls, name, bases, namespace, **kwargs):
        if abc.ABC not in bases:
            if not namespace.get("default_proxy"):
                raise exceptions.PluginError("You must define default_proxy")
            if not namespace.get("proxy_page_url"):
                raise exceptions.PluginError("You must define proxy_page_url")
        return type.__new__(cls, name, bases, namespace)


class ProxyPlugin(mixins.RequestMixin, abc.ABC, metaclass=ProxyMeta):
    """Get a list of proxies for given website."""

    log = logging.getLogger(__name__)
    default_proxy = ""  # To be overwritten by subclasses
    proxy_page_url = ""  # To be overwritten by subclasses

    async def get_proxy(self) -> Url:
        """Return the first working proxy.

        Returns:
            The first working proxy.
        """
        self.log.debug("Getting list of proxies.")
        try:
            proxy_page = await self.get(self.proxy_page_url)
            return self._parse(proxy_page)
        except (OSError, asyncio.TimeoutError, exceptions.NoProxyAvailable):
            return self.default_proxy

    @staticmethod
    @abc.abstractmethod
    def _parse(page: str) -> str:
        """Parse proxy page a yield a proxies.

        Args:
            page: Page with a list of proxies.

        Returns:
            A Proxy URL.
        """
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _is_online(soup: bs4.Tag) -> bool:
        """Check if listed proxy is online.

        Args:
            soup: BeautifulSoup for given proxy.

        Returns:
            True if site is listed online.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['mariner', 'mariner.commands', 'mariner.proxies', 'mariner.trackers']

package_data = \
{'': ['*'], 'mariner': ['config/*']}

install_requires = \
['aiodns>=1.1.1,<2.0.0',
 'aiofiles>=0.4.0,<0.5.0',
 'aiohttp>=3.5.1,<4.0.0',
 'bs4>=0.0.1,<0.0.2',
 'cachalot>=1.0,<2.0',
 'cliff==2.9.1',
 'cmd2==0.7.8',
 'colorama>=0.4.1,<0.5.0',
 'future-fstrings>=0.4.4,<0.5.0',
 'lxml>=4.2.5,<5.0.0',
 'maya>=0.6.0,<0.7.0',
 'ruamel.yaml>=0.15.82,<0.16.0',
 'tokenize-rt>=2.1.0,<3.0.0']

entry_points = \
{'console_scripts': ['mariner = mariner.main:main'],
 'mariner.cli': ['config = mariner.commands.config:Config',
                 'details = mariner.commands.details:Details',
                 'download = mariner.commands.download:Download',
                 'magnet = mariner.commands.magnet:Magnet',
                 'open = mariner.commands.open:Open',
                 'search = mariner.commands.search:Search']}

setup_kwargs = {
    'name': 'mariner',
    'version': '1.4.1',
    'description': 'Navigate torrents in CLI',
    'long_description': "# Mariner [![PyPI version](https://badge.fury.io/py/mariner.svg)](https://badge.fury.io/py/mariner) [![Pipeline status](https://gitlab.com/radek-sprta/mariner/badges/master/pipeline.svg)](https://gitlab.com/radek-sprta/mariner/commits/master) [![Coverage report](https://gitlab.com/radek-sprta/mariner/badges/master/coverage.svg)](https://gitlab.com/radek-sprta/mariner/commits/master) [![Downloads](http://pepy.tech/badge/mariner)](http://pepy.tech/project/mariner)\n\nNavigate torrents in CLI with Mariner. It offers a simple interface for streamlined experience. No more annoying ads and pop-up windows.\n\nIt is currently under heavy development, so expect breaking changes. Currently only works in Linux, but any contributions in this regard are welcome.\n\n## Features\n\n- Automatically get a working proxy for trackers that have them.\n- Download torrent files and copy magnet links to clipboard.\n- Open torrents in your default torrent application.\n- Show torrent details.\n- Asynchronous I/O for better responsiveness.\n- Supports the following trackers:\n  - Anidex\n  - Archive.org\n  - Distrowatch\n  - Etree\n  - KickAssTorrents\n  - LimeTorrents\n  - Linuxtracker\n  - Nyaa\n  - NyaaPantsu\n  - ThePirateBay\n  - TokyoTosho\n\n![Mariner demonstration](docs/assets/mariner.svg)\n\n## Installation\n\nMariner requires Python 3.5.3 or newer to run.\n\n### Python package\n\nYou can easily install Mariner using pip. This is the preferred method:\n\n`pip3 install mariner`\n\n### Manual\n\nAlternatively, to get the latest development version, you can clone this repository and then manually install it:\n\n``` bash\ngit clone git@gitlab.com:radek-sprta/mariner.git\ncd mariner\npoetry build\npip install dist/*.whl\n```\n\n## Usage\n\nMariner supports both interactive and non-interactive modes. To see the list of commands, simply type:\n\n`mariner help`\n\nIn order to start Mariner in interactive mode, run it without any arguments:\n\n`mariner`\n\nThen search for Ubuntu torrents:\n\n`(mariner) search Ubuntu -t linuxtracker`\n\nand download the first result on the list:\n\n`(mariner) download 0`\n\nAlternatively, copy the magnet link to clipboard:\n\n`(mariner) magnet 0`\n\nOr open it in your torrent application:\n\n`(mariner) open 0`\n\nAnd quit the program:\n\n`(mariner) quit`\n\nFor more information, check the [documentation][documentation].\n\n## TODO\n\n- Add more trackers for searching\n- Support for MacOS and Windows\n\n## Contributing\n\nFor information on how to contribute to the project, please check the [Contributor's Guide][contributing]\n\n## Disclaimer\n\nI do not encourage anyone to act in conflict with their local laws and I do not endorse any illegal activity. Some content in the search results provided be Mariner might be illegal in your country and it is up to you to check your local laws before using it. Neither I, nor Mariner can be held liable for any action taken against you as the result of using it.\n\n## Contact\n\n[mail@radeksprta.eu](mailto:mail@radeksprta.eu)\n\n[incoming+radek-sprta/mariner@gitlab.com](incoming+radek-sprta/mariner@gitlab.com)\n\n## Acknowledgements\n\nMariner uses many excellent open-source libraries. But I would particularly like to mention the following, as without them, Mariner might not have been possible:\n\n- [Aiohttp](https://github.com/aio-libs/aiohttp)\n- [BeautifulSoup](https://www.crummy.com/software/BeautifulSoup/)\n- [Cliff](https://github.com/openstack/cliff/tree/master/cliff)\n- [TinyDB](https://github.com/msiemens/tinydb)\n\n## License\n\nGNU General Public License v3.0\n\n[contributing]: https://gitlab.com/radek-sprta/mariner/blob/master/CONTRIBUTING.md\n[documentation]: https://radek-sprta.gitlab.io/mariner\n",
    'author': 'Radek Sprta',
    'author_email': 'mail@radeksprta.eu',
    'url': 'https://radek-sprta.gitlab.io/mariner/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5.3,<4.0.0',
}


setup(**setup_kwargs)



"""Mariner - navigate torrents in CLI."""
VERSION = (1, 4, 1)

__title__ = "Mariner"
__description__ = "Navigate torrents in CLI"
__url__ = "http://gitlab.com/radek-sprta/mariner"
__download_url__ = "https://gitlab.com/radek-sprta/mariner/repository/archive.tar.gz?ref=master"
__version__ = ".".join(map(str, VERSION))
__author__ = "Radek Sprta"
__author_email__ = "mail@radeksprta.eu"
__license__ = "GNU GPLv3"
__copyright__ = "Copyright 2017 Radek Sprta"



# -*- coding: future_fstrings -*-
"""Mariner open command."""
import logging
import os
import shlex
import subprocess  # nosec

from cliff import command

from mariner import downloader


class Open(command.Command):
    """Open torrent in the default torrent application."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument("ID", nargs="+", help="ID of the torrents to open", type=int)
        return parser

    @staticmethod
    def _get_torrent_link(torrent_):
        """Get a link torrent, that can be opened by xdg-open.

        Args:
            torrent: Torrent object.

        Returns:
            Openable torrent link.
        """
        if torrent_.torrent:
            torrent_downloader = downloader.Downloader()
            torrent_downloader.download([(torrent_.torrent, torrent_.filename)])
            return str(torrent_downloader.download_path / torrent_.filename)
        return torrent_.magnet

    def take_action(self, parsed_args):
        """Open chosen torrent in the default torrent application.

        Args:
            parsed_args: List of parsed arguments.
        """
        for tid in parsed_args.ID:
            torrent_ = self.app.engine.result(tid)
            self.log.debug("tid=%s torrent=%s", tid, torrent_)
            self.log.info(f"Opening {torrent_.colored().name}.")
            link = shlex.quote(self._get_torrent_link(torrent_))
            if self.app.options.verbose_level > 1:
                subprocess.run(["xdg-open", link], check=False)
            else:
                with open(os.devnull) as devnull:
                    subprocess.run(["xdg-open", link], stdout=devnull, stderr=devnull, check=False)
            try:
                # Try deleting the file, if it exists
                link.unlink()
            except AttributeError:
                pass



# -*- coding: future_fstrings -*-
"""CLI config command."""
import logging
import pprint

from cliff import command

from mariner import exceptions, utils


class Config(command.Command):
    """Show or update configuration."""

    log = logging.getLogger(__name__)

    def get_parser(self, prog_name):
        """Add arguments to argument parser.

        Args:
            prog_name: Application name.

        Returns:
            Instance of argument parser.
        """
        parser = super().get_parser(prog_name)
        parser.add_argument("--show", "-s", action="store_true", help="Show the configuration")
        parser.add_argument("key", nargs="?", help="Option to change", default=None)
        parser.add_argument("value", nargs="?", help="New value", default=None)
        return parser

    @staticmethod
    def _update_dict(dict_, key, value):
        """Update option in dictionary to a new value.

        Args:
            dict_: Dictionary to update.
            key: Option to update.
            value: New value.
        """
        # If the key is in the dictionary, update the value
        if key in dict_:
            dict_[key] = value
            return
        # Otherwise, recursively look for the key
        for c in dict_.values():
            if not isinstance(c, dict):
                continue
            Config._update_dict(c, key, value)
        # If the key is not there, raise an error
        raise exceptions.InputError("Wrong configuration option")

    def take_action(self, parsed_args):
        """Copy chosen magnet link to clipboard.

        Args:
            parsed_args: List of parsed arguments.
        """
        key = parsed_args.key
        value = parsed_args.value
        show_ = parsed_args.show
        if not show_ and not (key and value):
            raise ValueError("Provide key and value to update or use --show")
        if show_:
            self.log.info("Mariner configuration:")
            pprint.pprint(self.app.config._config)  # pylint: disable=protected-access
        else:
            self._update_dict(self.app.config, key, value)
            self.log.info(f"Updated {utils.green(key)} to {utils.green(value)}")



# -*- coding: future_fstrings -*-
"""Module for searching torrents on Anidex."""
import logging
from typing import Iterator

import bs4

from mariner import torrent, trackerplugin


class Anidex(trackerplugin.TrackerPlugin):
    """Represents Anidex.info search engine."""

    log = logging.getLogger(__name__)

    filters = {"anime"}
    search_url = "https://anidex.info/?q={title}&s=seeders&o=desc"

    def _parse(self, raw: str) -> Iterator[torrent.Torrent]:  # pylint: disable=too-many-locals
        """Parse result page.

        Args:
          raw: Raw HTML results to parse.

        Returns:
            List of torrent names with magnet links and URLs.
        """
        soup = bs4.BeautifulSoup(raw, "lxml")
        try:
            contents = soup.select("table.table-striped tbody tr")
            for content in contents:
                data = content.select("td")
                name = str(data[2].select("span.span-1440")[0].string)
                tracker = self.__class__.__name__

                url_stub = data[4].select("a")[0].get("href")
                url = f"https://anidex.info{url_stub}"
                magnet = data[5].select("a")[0].get("href")

                size = str(data[6].string)
                date = str(data[7].string)
                seeds = self._parse_number(data[8].text)
                leeches = self._parse_number(data[9].string)

                yield torrent.Torrent(
                    name,
                    tracker,
                    magnet=magnet,
                    torrent=url,
                    size=size,
                    date=date,
                    seeds=seeds,
                    leeches=leeches,
                )
        except IndexError:
            yield from []
